/*
 * bmp2pbm.c
 *
 * BMP -> PGM & PPM conversion utility.
 *
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "pbm.h"

/* Mirror a pixel: BGR -> RGB */
#define MIRROR(bgr) (temp = bgr[0], bgr[0] = bgr[2], bgr[2] = temp)

void bmp2pbm(char *in_path, char *out_path)
{
  FILE *in_file, *out_file;
  BITMAPFILEHEADER bmp_hdr1;
  BITMAPINFOHEADER bmp_hdr2;
  int i, bmp_size, scanline_size, pixel_size, cur_pos, pad_bytes, rgb_val;
  int width, height;
  char *scanline, *next_pixel, temp;

  /* Open files */
  if (((in_file = fopen(in_path, "rb")) == NULL) || ((out_file = fopen(out_path, "wb")) == NULL)) {
    printf("Can`t open/create file(s).\n");
    exit(1);
  }

  bmp_size = file_size(in_file);

  /* Read BMP headers */
  fread(&bmp_hdr1, 1, 14, in_file);
  fread(&bmp_hdr2, 1, 40, in_file);

  /* Check BMP file */
  if ((bmp_hdr1.bfType != 0x4d42)          ||
     (bmp_hdr1.bfSize != (DWORD) bmp_size) ||
     (bmp_hdr2.biWidth > 16384)            ||
     (bmp_hdr2.biHeight > 16384)           ||
     (bmp_hdr2.biPlanes != 1)              ||
     (bmp_hdr2.biCompression != 0)         ||
     ((bmp_hdr2.biBitCount != 8) && (bmp_hdr2.biBitCount != 24)))
  {
    printf("Corrupted file or unsupportded format.\n");
    exit(1);
  }

  /* 8-bit grayscale BMP files must have a palette */ 
  if ((bmp_hdr1.bfOffBits != 14 + 1024 + bmp_hdr2.biSize) && (bmp_hdr2.biBitCount == 8)) {
    printf("8-bit BMP`s must have linear grayscale palette.\n");
    exit(1);
  }

  /* For forward compatibility (see MSDN) */
  fseek(in_file, 14 + bmp_hdr2.biSize, SEEK_SET);

  /* Only linear grayscale palette is acceptable */
  if (bmp_hdr2.biBitCount == 8) {
    for (i = 0; i < 256; i++) {
      fread(&rgb_val, 4, 1, in_file);
      if (rgb_val != ((i << 16) | (i << 8) | (i))) {
        printf("8-bit BMP`s must have linear grayscale palette.\n");
        exit(1);
      }
    }
  }

  width = bmp_hdr2.biWidth;
  height = bmp_hdr2.biHeight;
  pixel_size = (bmp_hdr2.biBitCount == 8 ? 1 : 3);

  /* Compute scanline size, including padding to 4-byte boundary */
  scanline_size = width * pixel_size;
  if (scanline_size & 3) scanline_size = (scanline_size & 0xfffffffc) + 4;

  pad_bytes = scanline_size - width * pixel_size;

  if ((scanline = (char *) malloc(scanline_size)) == NULL) {
    printf("Not enough memory.\n");
    exit(1);
  }

  /* Set BMP file pointer to the last scanline */
  fseek(in_file, bmp_size - scanline_size, SEEK_SET);

  /* Write PGM or PPM header */
  if (bmp_hdr2.biBitCount == 8)
  fprintf(out_file, "P5\n%d %d\n255\n", width, height);
  else fprintf(out_file, "P6\n%d %d\n255\n", width, height);

  /* Since BMP stores scanlines bottom-to-top, we have to invert the image
   * to PGM/PPM top-to-bottom order. To do this, we read the incoming data
   * in reverse order.
   */

  while (height > 0) {
    fread(scanline, 1, scanline_size, in_file);
    cur_pos = ftell(in_file);
    fseek(in_file, cur_pos - 2 * scanline_size, SEEK_SET);

    /* BMP stores pixels in reverse order: BGR instead of RGB */
    if (pixel_size == 3) {
      for (i = 0, next_pixel = scanline; i < width; i++) {
        MIRROR(next_pixel);
        next_pixel += 3;
      }
    }

    fwrite(scanline, 1, scanline_size - pad_bytes, out_file);
    height--;
  }

  fclose(in_file); fclose(out_file);
  free(scanline);
}

int main(int argc, char *argv[])
{
  printf("\nBMP -> PGM & PPM conversion utility, (C) ENTROPYWARE, Alexander Simakov.\n");

  if (argc == 3) {
    bmp2pbm(argv[1], argv[2]);
    printf("\n");
    return 0;
  }

  printf("\nUsage: <bmp2pbm.exe> <in_file> <out_file>\n\n");
  printf("Example 1: bmp2pbm.exe photo.bmp photo.pgm\n");
  printf("Example 2: bmp2pbm.exe flower.bmp flower.ppm\n");

  return 1;
}
