/*
 * gsbmp.c
 *
 * Dynamic Link Library for compressing BMP files using GSLib.
 *
 * (C) ENTROPYWARE, Alexander Simakov.
 *
 * Our site: http://www.entropyware.info/
 * Support:  support@entropyware.info
 * Author:   xander@entropyware.info
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "..\gslib.h"
#include "gsbmplib.h"

int file_size(FILE *f)
{
  int save_pos, size_of_file;

  save_pos = ftell(f);
  fseek(f, 0L, SEEK_END);
  size_of_file = ftell(f);
  fseek(f, save_pos, SEEK_SET);
  return size_of_file;
}

/* Memory allocation function */
void *GS_API my_alloc(int size, int items)
{
  return malloc(size * items);
}

/* Memory releasing function */
void GS_API my_free(void *address)
{
  free(address);
}

int GS_API gs_compress_bmp(char *in_path, char *out_path, int quality, void (GS_API *my_progress)(int percent_done))
{
  FILE *in_file, *out_file;
  char *in_buf, *out_buf;
  BITMAPFILEHEADER bmp_hdr1;
  BITMAPINFOHEADER bmp_hdr2;
  int bmp_size, pixel_size, scanline_size, pad_bytes, buf_size, cur_pos;
  int i, width, height, rgb_val, res;
  gs_stream gs_strm;

  /* Open BMP file */
  if ((in_file = fopen(in_path, "rb")) == NULL) {
    _fcloseall();
    return GSBMP_OPEN_ERROR;
  }

  /* Create GS file */
  if ((out_file = fopen(out_path, "wb")) == NULL) {
    _fcloseall();
    remove(out_path);
    return GSBMP_CREATE_ERROR;
  }

  bmp_size = file_size(in_file);

  /* Read BMP headers */
  fread(&bmp_hdr1, 1, 14, in_file);
  fread(&bmp_hdr2, 1, 40, in_file);

  /* Check BMP file */
  if ((bmp_hdr1.bfType != 0x4d42)          ||
     (bmp_hdr1.bfSize != (DWORD) bmp_size) ||
     (bmp_hdr2.biWidth > 16384)            ||
     (bmp_hdr2.biHeight > 16384)           ||
     (bmp_hdr2.biPlanes != 1)              ||
     (bmp_hdr2.biCompression != 0)         ||
     ((bmp_hdr2.biBitCount != 8) && (bmp_hdr2.biBitCount != 24)) ||
     ((bmp_hdr1.bfOffBits != 14 + 1024 + bmp_hdr2.biSize) && (bmp_hdr2.biBitCount == 8)))
  {
    _fcloseall();
    remove(out_path);
    return GSBMP_DAMAGED_BMP;
  }

  /* For forward compatibility (see MSDN) */
  fseek(in_file, 14 + bmp_hdr2.biSize, SEEK_SET);

  /* Only linear grayscale palette is acceptable */
  if (bmp_hdr2.biBitCount == 8) {
    for (i = 0; i < 256; i++) {
      fread(&rgb_val, 4, 1, in_file);
      if (rgb_val != ((i << 16) | (i << 8) | (i))) {
        _fcloseall();
        remove(out_path);
        return GSBMP_DAMAGED_BMP;
      }
    }
  }

  width = bmp_hdr2.biWidth;
  height = bmp_hdr2.biHeight;
  pixel_size = (bmp_hdr2.biBitCount == 8 ? 1 : 3);

  /* Compute scanline size, including padding to 4-byte boundary */
  scanline_size = width * pixel_size;
  if (scanline_size & 3) scanline_size = (scanline_size & 0xfffffffc) + 4;

  pad_bytes = scanline_size - width * pixel_size;

  /* Set buf_size to one scanline */
  buf_size = scanline_size;

  /* Allocate buffers */
  in_buf = (char *) malloc(buf_size);
  out_buf = (char *) malloc(buf_size);

  if (in_buf == NULL || out_buf == NULL) {
    _fcloseall();
    remove(out_path);
    return GSBMP_MEMORY_ERROR;
  }

  /* Prepare GSLib */
  gs_strm.next_in = in_buf;
  gs_strm.avail_in = 0;

  gs_strm.next_out = out_buf;
  gs_strm.avail_out = 0;

  gs_strm.img_width = width;
  gs_strm.img_height = height;

  gs_strm.quality = quality;

  gs_strm.gs_alloc = my_alloc;
  gs_strm.gs_free = my_free;
  gs_strm.gs_progress = my_progress;

  /* Start GSLib */
  if (pixel_size == 1) res = gs_start_compress(&gs_strm, GS_GRAY); /* 8-bit BMP */
  else res = gs_start_compress(&gs_strm, GS_BGR);                  /* 24-bit BMP */

  if (res != GS_OK) {
    _fcloseall();
    remove(out_path);
    return GSBMP_INTERNAL_ERROR;
  }

  /* Since BMP stores scanlines bottom-to-top, we have to invert the image
   * to GS's top-to-bottom order. To do this, we read the incoming data
   * in reverse order.
   */

  /* Set BMP file pointer to the last scanline */
  fseek(in_file, bmp_size - scanline_size, SEEK_SET);

  /* Main compression loop */

  do {

    /* Compress next portion of data */
    res = gs_compress(&gs_strm);

    switch (res) {

      /* Input buffer is empty, read more input */
      case GS_BUFFER_EMPTY:
      /* Read one scanline */
      fread(in_buf, 1, scanline_size, in_file);

      /* Set file pointer to the previous scanline */
      cur_pos = ftell(in_file);
      fseek(in_file, cur_pos - 2 * scanline_size, SEEK_SET);
      gs_strm.next_in = in_buf;
      gs_strm.avail_in = scanline_size - pad_bytes;
      break;

      /* Output buffer is full, write more output */
      case GS_BUFFER_FULL:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      gs_strm.next_out = out_buf;
      gs_strm.avail_out = buf_size;
      break;

      /* Work complete, flush remaining output */
      case GS_OK:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf, out_file);
      break;
    }
  } while (res != GS_OK);

  /* Stop GSLib */
  gs_end_compress(&gs_strm);

  _fcloseall();
  free(in_buf); free(out_buf);

  return GSBMP_OK;
}

int GS_API gs_decompress_bmp(char *in_path, char *out_path, void (GS_API *my_progress)(int percent_done))
{
  FILE *in_file, *out_file;
  char *in_buf = NULL, *out_buf = NULL;
  BITMAPFILEHEADER bmp_hdr1;
  BITMAPINFOHEADER bmp_hdr2;
  int gs_size, bmp_size, scanline_size = 0, buf_size = 0, pad_bytes = 0, bytes_read, cur_pos;
  int i, rgb_val, res;
  gs_stream gs_strm;

  /* Open GS file */
  if ((in_file = fopen(in_path, "rb")) == NULL) {
    _fcloseall();
    return GSBMP_OPEN_ERROR;
  }

  /* Create BMP file */
  if ((out_file = fopen(out_path, "wb")) == NULL) {
    _fcloseall();
    remove(out_path);
    return GSBMP_CREATE_ERROR;
  }

  gs_size = file_size(in_file);

  /* Allocate input buffer */
  if ((in_buf = (char *) malloc(BUFFER_SIZE)) == NULL) {
    _fcloseall();
    remove(out_path);
    return GSBMP_MEMORY_ERROR;
  }

  /* Prepare GSLib */
  gs_strm.next_in = in_buf;
  gs_strm.avail_in = 0;

  gs_strm.gs_alloc = my_alloc;
  gs_strm.gs_free = my_free;
  gs_strm.gs_progress = my_progress;

  /* Start GSLib */
  res = gs_start_decompress(&gs_strm, gs_size);

  if (res != GS_OK) {
    _fcloseall();
    remove(out_path);
    return GSBMP_INTERNAL_ERROR;
  }

  /* Main decompression loop */

  do {

    /* Decompress next portion of data */
    res = gs_decompress(&gs_strm);

    switch (res) {

      /* Input buffer is empty, read more input */
      case GS_BUFFER_EMPTY:
      bytes_read = fread(in_buf, 1, BUFFER_SIZE, in_file);
      gs_strm.next_in = in_buf;
      gs_strm.avail_in = bytes_read;
      break;

      /* Output buffer is full, write more output */
      case GS_BUFFER_FULL:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf + pad_bytes, out_file);
      cur_pos = ftell(out_file);
      fseek(out_file, cur_pos - 2 * scanline_size, SEEK_SET);
      gs_strm.next_out = out_buf;
      gs_strm.avail_out = buf_size - pad_bytes;
      break;

      /* GS header is damaged */
      case GS_HEADER_FAIL:
      _fcloseall();
      remove(out_path);
      return GSBMP_DAMAGED_GS;
      break;

      /* Not enough memory */
      case GS_MEMORY_FAIL:
      _fcloseall();
      remove(out_path);
      return GSBMP_MEMORY_ERROR;
      break;

      /* GS header is successfully read, write BMP header */
      case GS_HEADER_OK:

      /* Compute scanline size, including padding to 4-byte boundary */
      scanline_size = gs_strm.img_width;
      if (scanline_size & 3) scanline_size = (scanline_size & 0xfffffffc) + 4;

      buf_size = scanline_size;

      pad_bytes = scanline_size - gs_strm.img_width;

      /* Allocate output buffer */
      if ((out_buf = (char *) malloc(buf_size)) == NULL) {
        _fcloseall();
        remove(out_path);
        return GSBMP_MEMORY_ERROR;
      }

      gs_strm.next_out = out_buf;
      gs_strm.avail_out = buf_size - pad_bytes;

      /* Compute BMP file size */
      bmp_size = 14 + 40 + 1024 + scanline_size * gs_strm.img_height;

      /* Fill BITMAPFILEHEADER */
      bmp_hdr1.bfType = 0x4d42;
      bmp_hdr1.bfSize = bmp_size;
      bmp_hdr1.bfReserved1 = 0;
      bmp_hdr1.bfReserved2 = 0;
      bmp_hdr1.bfOffBits = 14 + 40 + 1024;

      /* Fill BITMAPINFOHEADER */
      bmp_hdr2.biSize = 40;
      bmp_hdr2.biWidth = gs_strm.img_width;
      bmp_hdr2.biHeight = gs_strm.img_height;
      bmp_hdr2.biPlanes = 1;
      bmp_hdr2.biBitCount = 8;
      bmp_hdr2.biCompression = 0;
      bmp_hdr2.biSizeImage = 0;
      bmp_hdr2.biXPelsPerMeter = 0;
      bmp_hdr2.biYPelsPerMeter = 0;
      bmp_hdr2.biClrUsed = 0;
      bmp_hdr2.biClrImportant = 0;

      /* Write BMP headers */
      fwrite(&bmp_hdr1, 1, 14, out_file);
      fwrite(&bmp_hdr2, 1, 40, out_file);

      /* Write linear grayscale palette */
      for (i = 0; i < 256; i++) {
        rgb_val = ((i << 16) | (i << 8) | (i));
        fwrite(&rgb_val, 1, 4, out_file);
      }

      fseek(out_file, bmp_size - scanline_size, SEEK_SET);
      break;

      /* Work complete, flush remaining output */
      case GS_OK:
      fwrite(out_buf, 1, gs_strm.next_out - out_buf + pad_bytes, out_file);
      break;
    }
  } while (res != GS_OK);

  /* Stop GSLib */
  gs_end_decompress(&gs_strm);

  free(in_buf); free(out_buf);
  _fcloseall();

  return GSBMP_OK;
}

int GS_API gsbmp_main(int x, int y, int z) {
  x = y = z; /* just to make compiler happy */
  return 1;
}
